var g_text_tid = mx.read_texture("@jstest/test.seq");
var g_digits_tid = mx.read_texture("@jstest/digits.seq");
var g_current_display = "";
var g_current_digits = "";
var g_debug = false;

function break_time(t) {
	var min, sec, ms;

	ms = Math.floor(t * 1000.0);
	sec = Math.floor(ms / 1000);
	min = Math.floor(sec / 60);

	ms -= sec * 1000;
	sec -= min * 60;

	return { min: min, sec: sec, ms: ms };
}

function left_fill_string(s, pad, n) {

	n -= s.length;

	while (n > 0) {
		if (n & 1)
			s = pad + s;

		n >>= 1;
		pad += pad;
	}

	return s;
}

function time_to_string(t) {
	var s;

	t = break_time(t);

	s = left_fill_string(t.min.toString(), " ", 0) + ":";
	s += left_fill_string(t.sec.toString(), "0", 2) + ".";
	s += left_fill_string(t.ms.toString(), "0", 3);

	return s;
}

function show_text(s) {
	var char_cells_across = 16;
	var char_cells_down = 8;
	var char_cell_width = 1 / char_cells_across / 2;
	var char_cell_height = 1 / char_cells_down / 2;
	var i, n, sx, sy, dx, dy;

	if (s == g_current_display)
		return;

	g_current_display = s;

	if (g_debug)
		mx.message("Updating texture to '" + s + "'");

	mx.begin_custom_frame(g_text_tid);
	dx = 0;
	dy = 0;
	for (i = 0; i < s.length; i++) {
		if (s[i] == "\n") {
			dx = 0;
			dy += char_cell_height;
			continue;
		}
		n = s.charCodeAt(i);
		sx = (n % char_cells_across) * char_cell_width;
		sy = (Math.floor(n / char_cells_across)) * char_cell_height;
		mx.paste_custom_frame(g_text_tid, 0, sx, sy, dx, dy, char_cell_width, char_cell_height);
		dx += char_cell_width;
	}
	mx.end_custom_frame(g_text_tid);
}

function show_digits(s) {
	var char_cells_across = 6;
	var char_cells_down = 3;
	var char_cell_width = Math.floor(256 / char_cells_across) / 256;
	var char_cell_height = Math.floor(256 / char_cells_down) / 256;
	var i, n, sx, sy, dx, dy;

	if (s == g_current_digits)
		return;

	g_current_digits = s;

	if (g_debug)
		mx.message("Updating texture to '" + s + "'");

	mx.begin_custom_frame(g_digits_tid);
	dx = 0;
	dy = 0;
	for (i = 0; i < s.length; i++) {
		if (s[i] == "\n") {
			dx = 0;
			dy += char_cell_height;
			continue;
		}
		if (s[i] == " ") {
			dx += char_cell_width;
			continue;
		}
		n = s.charCodeAt(i) - "0".charCodeAt(0);
		sx = (n % char_cells_across) * char_cell_width;
		sy = (Math.floor(n / char_cells_across)) * char_cell_height;
		mx.paste_custom_frame(g_digits_tid, 0, sx, sy, dx, dy, char_cell_width, char_cell_height);
		dx += char_cell_width;
	}
	mx.end_custom_frame(g_digits_tid);
}

var g_leader_slot = -1;

function show_lead_change(slot) {
	var a;

	a = [
		"New leader!",
		mx.get_rider_name(slot)
	];

	show_text(a.join("\n"));
}

function show_running_order() {
	var i, a, r;

	r = mx.get_running_order();

	if (r.length >= 1 && r[0].slot != g_leader_slot && r[0].position > 0) {
		show_lead_change(r[0].slot);
		g_leader_slot = r[0].slot;
		return;
	}

	a = [ "Running Order" ];
	for (i = 0; i < r.length; i++)
		a.push(left_fill_string((i + 1).toFixed(0), " ", 3) + " "
		 + left_fill_string(mx.get_rider_number(r[i].slot), " ", 3) + " "
		 + mx.get_rider_name(r[i].slot)
		 );

	show_text(a.join("\n"));
}

function show_tower_display() {
	var i, s, r;

	r = mx.get_running_order();

	if (r.length == 0)
		return;

	/* first number is laps */
	s = left_fill_string((mx.index_to_lap(r[0].position) + 1).toFixed(0), " ", 3);

	/* followed by rider numbers */
	for (i = 0; i < r.length; i++) {
		s += left_fill_string(mx.get_rider_number(r[i].slot), " ", 3);
		if ((i & 1) == 0)
			s += "\n";
	}

	show_digits(s);
}

var g_best_lap_slot = -1;
var g_best_lap_time = -1;
var g_lap_times = []; /* array of arrays, first index is slot, second is lap times starting from 2nd lap */
var g_best_laps = [];

function get_laps(slot) {
	return mx.index_to_lap(mx.get_timing_position(slot));
}

function get_lap_time(slot, lap) {
	var t0, t1;

	t0 = mx.get_timing(slot, mx.lap_to_index(lap - 1));
	t1 = mx.get_timing(slot, mx.lap_to_index(lap));

	return t1 - t0;
}

function update_lap_times_for_slot(slot) {
	var i, n, a, t, best, best_time;

	if (slot in g_lap_times)
		a = g_lap_times[slot];
	else
		a = [];

	n = get_laps(slot);

	if (n <= 1)
		return;

	if (slot in g_best_laps)
		best = g_best_laps[slot];
	else
		best = -1

	if (best >= 2)
		best_time = a[best - 2];
	else
		best_time = -1;

	for (i = a.length; i < n - 2; i++) {
		t = get_lap_time(slot, i + 2);
		if (best_time < 0 || t < best_time) {
			best = i + 2;
			best_time = t;

			if (g_best_lap_time < 0 || best_time < g_best_lap_time) {
				g_best_lap_slot = slot;
				g_best_lap_time = best_time
			}
		}
		a.push(t);
	}

	g_lap_times[slot] = a;
	g_best_laps[slot] = best;
}

function best_lap_time_for_slot(slot) {
	if (slot in g_best_laps)
		return g_lap_times[slot][g_best_laps[slot] - 2];
	return null;
}

function show_best_lap_change() {
	var a, slot, t;

	slot = g_best_lap_slot;
	t = best_lap_time_for_slot(slot);

	a = [
		"New fast lap!",
		time_to_string(t),
		mx.get_rider_name(slot)
	];

	show_text(a.join("\n"));
}

function show_best_lap() {
	var i, r, a, t, e, prev_best;

	r = mx.get_running_order();

	a = [];

	try {
		prev_best = g_best_lap_time;

		for (i = 0; i < r.length; i++)
			update_lap_times_for_slot(r[i].slot);

		if (g_best_lap_time < prev_best) {
			show_best_lap_change();
			return true;
		}

		for (i = 0; i < r.length; i++) {
			t = best_lap_time_for_slot(r[i].slot);
			if (t)
				a.push(time_to_string(t) + " " + mx.get_rider_name(r[i].slot));
		}

		if (a.length == 0)
			return false;

		a.sort();

		show_text("Best Laps\n" + a.join("\n"));
	} catch (e) {
		mx.message(e.toString());
	}

	return true;
}

function rewind(seconds) {
	g_leader_slot = -1;
	g_best_lap_slot = -1;
	g_best_lap_time = -1;
	g_lap_times = [];
	g_best_laps = [];
}

var g_last_update = -10;

function pose(seconds) {
	var statue_index = 0;
	var bone_count = 2;
	var bone_centers = [ 0, 0, 0, 0, 0, 0 ];
	var bone_positions = [ 0, 0, 0, 0, 0, 0 ];
	var a = Math.sin(seconds * 10) * .5;
	var bone_rotations = [
	 1, 0, 0,
	 0, 1, 0,
	 0, 0, 1,
	 Math.cos(a), 0, -Math.sin(a),
	 0, 1, 0,
	 Math.sin(a), 0, Math.cos(a)
	];

	mx.pose_statue(statue_index, bone_count, bone_centers, bone_positions, bone_rotations);
}

mx.frame_handler = function (seconds) {
	var l, s;

	pose(seconds);

	if (seconds < g_last_update) {
		rewind(seconds);
		g_last_update = 0;
	}

	l = Math.floor(g_last_update / 5);
	s = Math.floor(seconds / 5);

	if (l == s)
		return;

	g_last_update = seconds;

	try {
		show_tower_display();
	} catch (e) {
		mx.message(e);
	}

	if (s % 2 == 0)
		show_running_order();
	else
		show_best_lap() || show_running_order();
}

show_text("READY");
